# mypy: ignore-errors
# pylint: skip-file
"""
Test the labelset module.
"""

from typing import Any
from unittest.mock import patch
import shutil
import json
import os
import matplotlib.pyplot as plt
import numpy as np

import redbrick
from redbrick.entity.custom_group import CustomGroup
from .taxonomy2_test import TAXONOMY1
from .video_test import VIDEO, VIDEO_CLASSIFY, VIDEO_BBOX_REAL
from .image_test import IMAGE


# d888888b .88b  d88.  .d8b.   d888b  d88888b
#   `88'   88'YbdP`88 d8' `8b 88' Y8b 88'
#    88    88  88  88 88ooo88 88      88ooooo
#    88    88  88  88 88~~~88 88  ooo 88~~~~~
#   .88.   88  88  88 88   88 88. ~8~ 88.
# Y888888P YP  YP  YP YP   YP  Y888P  Y88888P


class Mock_Image:
    def __init__(self):
        return

    def get_datapoint_ids(self, org_id, label_set_name):
        return (
            ["1", "2"],
            CustomGroup("BBOX", "IMAGE", TAXONOMY1["taxonomy"]),
        )

    def get_datapoint(self, orgid, labelsetname, dpid, tasktype, taxonomy):
        return IMAGE


@patch("redbrick.labelset.loader.RedBrickApi")
def test_labelset_init(mock: Any) -> None:
    """Test init of labelset."""

    mock.return_value = Mock_Image()
    labelset = redbrick.labelset.LabelsetLoader(org_id="123", label_set_name="abc")
    assert labelset.task_type == "BBOX"
    assert len(labelset.dp_ids) == 2


@patch("redbrick.labelset.loader.RedBrickApi")
def test_labelset_export_image(mock: Any) -> None:
    """Test labelset export for images."""
    mock.return_value = Mock_Image()

    labelset = redbrick.labelset.LabelsetLoader(org_id="123", label_set_name="abc")
    cache_dir = labelset.export(format="redbrick")

    assert os.path.isdir(cache_dir)
    assert os.listdir(os.path.join(cache_dir, "obj_train_data")).sort() == [
        IMAGE.image_url_not_signed[0:-4] + ".png",
        IMAGE.image_url_not_signed[0:-4] + ".txt",
    ].sort()
    shutil.rmtree(cache_dir)


# db    db d888888b d8888b. d88888b  .d88b.
# 88    88   `88'   88  `8D 88'     .8P  Y8.
# Y8    8P    88    88   88 88ooooo 88    88
# `8b  d8'    88    88   88 88~~~~~ 88    88
#  `8bd8'    .88.   88  .8D 88.     `8b  d8'
#    YP    Y888888P Y8888D' Y88888P  `Y88P'


class Mock_Video:
    """Mocking RedBrickApi for video."""

    def __init__(self, data=VIDEO, tasktype="BBOX", num_pts=2):
        self.data = data
        self.task_type = tasktype
        self.dps = [str(i) for i in range(num_pts)]

    def get_datapoint_ids(self, org_id, label_set_name):
        return (
            self.dps,
            CustomGroup(self.task_type, "VIDEO", TAXONOMY1["taxonomy"]),
        )

    def get_datapoint(self, orgid, labelsetname, dpid, tasktype, taxonomy):
        if tasktype == "CLASSIFY":
            self.data.video_name = dpid
        return self.data


@patch("redbrick.export.export_video.url_to_image")
@patch("redbrick.labelset.loader.RedBrickApi")
def test_labelset_export_video(mock: Any, url_to_image: Any) -> None:
    """Test labelset export for images."""
    mock.return_value = Mock_Video()
    url_to_image.return_value = np.zeros((10, 10, 3))

    labelset = redbrick.labelset.LabelsetLoader(org_id="123", label_set_name="abc")
    cache_dir = labelset.export(format="redbrick")

    assert os.path.isdir(cache_dir)

    outdir = []
    for item in VIDEO.items_list_not_signed:
        outdir += [(item[0:-4] + ".txt").replace("/", "_").replace(":", "_")]
        outdir += [(item[0:-4] + ".png").replace("/", "_").replace(":", "_")]

    assert os.listdir(os.path.join(cache_dir, "obj_train_data")).sort() == outdir.sort()
    shutil.rmtree(cache_dir)


@patch("redbrick.export.export_video.url_to_image")
@patch("redbrick.labelset.loader.RedBrickApi")
def test_labelset_export_video_clasify(mock: Any, url_to_image: Any) -> None:
    """Test labelset export for images."""
    mock.return_value = Mock_Video(data=VIDEO_CLASSIFY, tasktype="CLASSIFY")
    url_to_image.return_value = np.zeros((10, 10, 3))

    labelset = redbrick.labelset.LabelsetLoader(org_id="123", label_set_name="abc")
    cache_dir = labelset.export(format="redbrick")

    assert os.path.isdir(cache_dir)

    files = os.listdir(cache_dir)
    assert len(files) == 2
    assert files.sort() == mock.return_value.dps.sort()
    shutil.rmtree(cache_dir)


@patch("redbrick.export.export_video.url_to_image")
@patch("redbrick.labelset.loader.RedBrickApi")
def test_labelset_export_video_clasify_json(mock: Any, url_to_image: Any) -> None:
    """Test labelset export for images."""
    mock.return_value = Mock_Video(data=VIDEO_CLASSIFY, tasktype="CLASSIFY")
    url_to_image.return_value = np.zeros((10, 10, 3))

    labelset = redbrick.labelset.LabelsetLoader(org_id="123", label_set_name="abc")
    cache_dir = labelset.export(format="redbrick-json")

    assert os.path.isdir(cache_dir)

    files = os.listdir(cache_dir)
    assert len(files) == 1

    # open the json file
    with open(os.path.join(cache_dir, files[0]), "r") as file:
        labels = json.load(file)

    shutil.rmtree(cache_dir)


@patch("redbrick.entity.datapoint.video.url_to_image")
@patch("redbrick.labelset.loader.RedBrickApi")
def test_labelset_showdata_video(mock: Any, url_to_image: Any) -> None:
    """Test labelset export for images."""
    mock.return_value = Mock_Video(data=VIDEO_BBOX_REAL, num_pts=20)
    url_to_image.return_value = np.ones((10, 10, 3)) * 0.5

    labelset = redbrick.labelset.LabelsetLoader(org_id="123", label_set_name="abc")

    plt.ion()  # make plt.show non blocking
    labelset.show_data()
    plt.close()



# mypy: ignore-errors
"""Test segmentation export functionality."""

from typing import Any
import filecmp
from unittest.mock import patch
import shutil
import os
import numpy as np  # type: ignore
import cv2

import redbrick
from tests.segmentation_test import remote_label5

from redbrick.entity.custom_group import CustomGroup
from redbrick.entity.datapoint import Image

# d8888b.  .d8b.  d888888b  .d8b.
# 88  `8D d8' `8b `~~88~~' d8' `8b
# 88   88 88ooo88    88    88ooo88
# 88   88 88~~~88    88    88~~~88
# 88  .8D 88   88    88    88   88
# Y8888D' YP   YP    YP    YP   YP

# Taxonomy
TAX = {
    "name": "DEFAULT::YoloThings",
    "version": 1,
    "categories": [
        {
            "name": "object",
            "children": [
                {"name": "person", "classId": 0, "children": []},
                {"name": "bicycle", "classId": 1, "children": []},
                {"name": "car", "classId": 2, "children": []},
                {"name": "motorbike", "classId": 3, "children": []},
                {"name": "aeroplane", "classId": 4, "children": []},
            ],
        }
    ],
}

# Image object
IMAGE = Image(
    org_id="123",
    label_set_name="abc",
    taxonomy={"person": 2, "car": 1, "background": 0},
    task_type="SEGMENTATION",
    remote_labels=remote_label5,
    dp_id="123",
    image_url="abc.png",
    image_url_not_signed="abc234.png",
    image_data=np.zeros((100, 100, 3)),
    created_by="123",
)

# Segmentation labelset mock
class MockImageSegmentation:
    """Image segmentation API mock."""

    def __init__(self) -> None:
        """Constructor."""
        return

    def get_datapoint_ids(self, org_id, label_set_name) -> Any:
        """Get data points ids."""
        return (
            ["1"],
            CustomGroup("SEGMENTATION", "IMAGE", TAX),
        )

    def get_datapoint(self, orgid, labelsetname, dpid, tasktype, taxonomy) -> Any:
        """Get the actual datapoint."""
        return IMAGE

    def get_members(self, org_id) -> Any:
        """get the members in the org."""
        x = {"123": "abc@def"}
        return x


@patch("redbrick.labelset.loader.RedBrickApi")
def test_export(mock: Any) -> None:
    """Test export."""
    # arrange
    mock.return_value = MockImageSegmentation()
    labelset = redbrick.labelset.LabelsetLoader(org_id="123", label_set_name="abc")

    # action
    cache_dir = labelset.export(format="redbrick")

    # assert
    expected_output = np.zeros([100, 100])
    expected_output[10:20, 10:20] = 1
    expected_output[12:18, 12:18] = 0
    expected_output[14:16, 14:16] = 2
    expected_output.dump("temp_test.dat")

    assert filecmp.cmp(
        "temp_test.dat", os.path.join(cache_dir, "abc234.png__export.dat")
    )

    # clean up
    shutil.rmtree(cache_dir)
    os.remove("temp_test.dat")


@patch("redbrick.labelset.loader.RedBrickApi")
def test_export_png(mock: Any) -> None:
    """Test export PNG Format."""
    # arrange
    mock.return_value = MockImageSegmentation()
    labelset = redbrick.labelset.LabelsetLoader(org_id="123", label_set_name="abc")

    print(
        "asdlfknadslkfnasldkfnalksdnfalskdnflkasdnflkasdnflkasdnflkasndflknsadklfnasldknflasdknfklasdnfklsadnfa",
        labelset.users,
    )
    # action
    cache_dir = labelset.export(format="redbrick-png")

    # assert
    expected_output = np.zeros([100, 100, 3])
    cmap = labelset[0].labels.color_map
    expected_output[10:20, 10:20, :] = cmap(1)
    expected_output[12:18, 12:18, :] = [0, 0, 0]
    expected_output[14:16, 14:16,] = cmap(2)
    # pylint: disable=no-member
    cv2.imwrite("temp_test.png", np.flip(expected_output, axis=2))

    assert filecmp.cmp(
        "temp_test.png", os.path.join(cache_dir, "abc234.png__export.png")
    )

    # clean up
    shutil.rmtree(cache_dir)
    os.remove("temp_test.png")



"""Public interface to remote_label."""
from typing import List, Union
from termcolor import colored
from redbrick.api import RedBrickApi
from redbrick.entity.task import Task
from redbrick.entity.label import ImageBoundingBox, VideoBoundingBox, VideoClassify
from redbrick.entity.taxonomy2 import Taxonomy2


class RemoteLabel:
    """An interface to RemoteLabel brick."""

    def __init__(self, org_id: str, project_id: str, stage_name: str) -> None:
        """Construct RemoteLabel instance."""
        print(colored("[INFO]:", "blue"), "Initializing remote-labeling module...")
        self.org_id = org_id
        self.project_id = project_id
        self.stage_name = stage_name
        self.api_client = RedBrickApi(cache=False)

        # Gather stage information and store
        stage_info = self.api_client.get_stage(
            org_id=org_id, project_id=project_id, stage_name=stage_name
        )
        taxonomy = self.api_client.get_taxonomy(
            orgId=org_id,
            name=stage_info["outputTaxonomyName"],
            version=stage_info["outputTaxonomyVersion"],
        )
        self.taxonomy: Taxonomy2 = taxonomy
        self.task_type = stage_info["outputType"]

    def get_num_tasks(self) -> int:
        """Get the number of tasks queued."""
        num = self.api_client.get_num_remote_tasks(
            org_id=self.org_id, project_id=self.project_id, stage_name=self.stage_name
        )
        return num

    def get_task(self, num_tasks: int) -> List[Task]:
        """User facing function to get task."""
        print(colored("[INFO]:", "blue"), "Retrieving task from backend...", end=" ")
        task = self.__get_remote_labeling_task(num_tasks=num_tasks)
        if len(task) == 0:
            print(colored("\n[WARNING]:", "yellow"), "No more tasks in this stage.")
            return task
        print(colored("Done.", "green"))
        return task

    def submit_task(
        self,
        task: Task,
        labels: Union[ImageBoundingBox, VideoBoundingBox, VideoClassify],
    ) -> None:
        """User facing funciton to submit a task."""
        print(colored("[INFO]", "blue"), "Submitting task to backend...", end=" ")
        new_subname = "remote-labeling"

        # Check that label category matches taxonomy
        check, classname = labels.compare_taxonomy(taxonomy=self.taxonomy)
        if not check:
            raise ValueError(
                "%s is not a valid category for taxonomy %s"
                % (classname, self.taxonomy.name)
            )

        # Check label type
        if task.task_data_type == "IMAGE_BBOX":
            if not isinstance(labels, ImageBoundingBox):
                raise ValueError("Labels must be of type ImageBoundingBox!")
        if task.task_data_type == "VIDEO_BBOX":
            if not isinstance(labels, VideoBoundingBox):
                raise ValueError("Labels must be of type VideoBoundingBox!")

        # Put task data
        self.__put_task_data(
            dp_id=task.dp_id,
            sub_name=new_subname,
            task_data=labels,
            taxonomy_name=task.taxonomy["name"],
            taxonomy_version=task.taxonomy["version"],
            td_type=task.task_data_type,
        )

        # Put remote labeling task
        submit_task = Task(
            org_id=task.org_id,
            project_id=task.project_id,
            stage_name=task.stage_name,
            task_id=task.task_id,
            dp_id=task.dp_id,
            sub_name=new_subname,
            taxonomy=task.taxonomy,
            items_list=task.items_list,
            items_list_presigned=task.items_list_presigned,
            task_data_type=task.task_data_type,
        )
        self.__put_remote_labeling_task(submit_task)

        print(colored("Done.", "green"))

    def __put_task_data(
        self,
        dp_id: str,
        sub_name: str,
        task_data: Union[ImageBoundingBox, VideoBoundingBox, VideoClassify],
        taxonomy_name: str,
        taxonomy_version: str,
        td_type: str,
    ) -> None:
        """Read labels from local folder, and submit the labels."""
        task_datas = str(task_data)  # Stringify the json object

        self.api_client.putTaskData(
            org_id=self.org_id,
            project_id=self.project_id,
            dp_id=dp_id,
            stage_name=self.stage_name,
            sub_name=sub_name,
            task_data=task_datas,
            taxonomy_name=taxonomy_name,
            taxonomy_version=taxonomy_version,
            td_type=td_type,
        )

    def __put_remote_labeling_task(self, task: Task) -> None:
        """Put the remote labeling task to the backend."""
        finished_task = {
            "orgId": task.org_id,
            "projectId": task.project_id,
            "stageName": task.stage_name,
            "taskId": task.task_id,
            "newSubName": task.sub_name,
        }
        self.api_client.putRemoteLabelingTask(finished_task)

    def __get_remote_labeling_task(self, num_tasks: int) -> List[Task]:
        """Get the labeling tasks from API."""
        task = self.api_client.tasksToLabelRemote(
            orgId=self.org_id,
            projectId=self.project_id,
            stageName=self.stage_name,
            numTasks=num_tasks,
        )
        return task



"""
Testing the segmentation module.
"""

from redbrick.entity.label import ImageSegmentation
from redbrick.entity.label.segmentation import ImageSegmentationRemoteLabel
import matplotlib.pyplot as plt

# d8888b.  .d8b.  d888888b  .d8b.
# 88  `8D d8' `8b `~~88~~' d8' `8b
# 88   88 88ooo88    88    88ooo88
# 88   88 88~~~88    88    88~~~88
# 88  .8D 88   88    88    88   88
# Y8888D' YP   YP    YP    YP   YP

remote_label1 = [
    {
        "category": [["object", "car"]],
        "attributes": [],
        "pixel": {"imagesize": [100, 100], "regions": [], "holes": [],},
        "labelid": "123",
    }
]

classes1 = {"car": 1, "person": 2}

remote_label2 = [
    {
        "category": [["object", "car"]],
        "attributes": [],
        "pixel": {
            "imagesize": [100, 100],
            "regions": [[[0, 0], [10, 0], [10, 10], [0, 10]]],
            "holes": [],
        },
        "labelid": "123",
    }
]

remote_label3 = [
    {
        "category": [["object", "car"]],
        "attributes": [],
        "pixel": {
            "imagesize": [100, 100],
            "regions": [[[0, 0], [10, 0], [10, 10], [0, 10]]],
            "holes": [],
        },
        "labelid": "123",
    },
    {
        "category": [["object", "person"]],
        "attributes": [],
        "pixel": {
            "imagesize": [100, 100],
            "regions": [[[9, 0], [19, 0], [19, 10], [9, 10]]],
            "holes": [],
        },
        "labelid": "123",
    },
]

remote_label4 = [
    {
        "category": [["object", "car"]],
        "attributes": [],
        "pixel": {
            "imagesize": [100, 100],
            "regions": [[[0, 0], [10, 0], [10, 10], [0, 10]]],
            "holes": [[[3, 3], [11, 3], [11, 11], [3, 11]]],
        },
        "labelid": "123",
    }
]

remote_label5 = [
    {
        "category": [["object", "person"]],
        "attributes": [],
        "pixel": {
            "imagesize": [100, 100],
            "regions": [[[14, 14], [16, 14], [16, 16], [14, 16]]],
            "holes": [],
        },
        "labelid": "123",
    },
    {
        "category": [["object", "car"]],
        "attributes": [],
        "pixel": {
            "imagesize": [100, 100],
            "regions": [[[10, 10], [20, 10], [20, 20], [10, 20]]],
            "holes": [[[12, 12], [18, 12], [18, 18], [12, 18]]],
        },
        "labelid": "123",
    },
]

# d888888b d88888b .d8888. d888888b .d8888.
# `~~88~~' 88'     88'  YP `~~88~~' 88'  YP
#    88    88ooooo `8bo.      88    `8bo.
#    88    88~~~~~   `Y8b.    88      `Y8b.
#    88    88.     db   8D    88    db   8D
#    YP    Y88888P `8888Y'    YP    `8888Y'


def test_init_empty() -> None:
    """Test basic initialization."""
    # arrange
    remote_label_1 = [
        ImageSegmentationRemoteLabel.from_dict(label) for label in remote_label1
    ]

    # action
    image_segment = ImageSegmentation(classes=classes1, remote_labels=remote_label_1)

    # assert
    assert image_segment.mask.shape == (100, 100)
    assert (image_segment.mask == 0).all()


def test_init_basic() -> None:
    """Test basic non empty initialization."""
    # arrange
    remote_label_2 = [
        ImageSegmentationRemoteLabel.from_dict(label) for label in remote_label2
    ]

    # action
    image_segment = ImageSegmentation(classes=classes1, remote_labels=remote_label_2)

    # assert
    assert (image_segment.mask[0:10, 0:10] == 1).all()
    assert (image_segment.mask[10:100, 10:100] == 0).all()


def test_init_two_obj() -> None:
    """Test basic init for multiple overlapping object."""
    # arrange
    remote_label_3 = [
        ImageSegmentationRemoteLabel.from_dict(label) for label in remote_label3
    ]

    # action
    image_segment = ImageSegmentation(classes=classes1, remote_labels=remote_label_3)

    # assert
    assert (image_segment.mask[0:9, 0:9] == 1).all()
    assert (image_segment.mask[0:9, 9:19] == 2).all()


def test_init_single_hole() -> None:
    """
    Test basic init for single hole single region. 
    This hole "spills outside" the region.
    """
    # arrange
    remote_label_4 = [
        ImageSegmentationRemoteLabel.from_dict(label) for label in remote_label4
    ]

    # action
    image_segment = ImageSegmentation(classes=classes1, remote_labels=remote_label_4)

    # assert
    assert (image_segment.mask[3:11, 3:11] == 0).all()


def test_init_reg_inside_hole() -> None:
    """Test a region inside another region's hole."""
    # arrange
    remote_label_5 = [
        ImageSegmentationRemoteLabel.from_dict(label) for label in remote_label5
    ]

    # action
    image_segment = ImageSegmentation(classes=classes1, remote_labels=remote_label_5)

    # assert
    assert (image_segment.mask[14:16, 14:16] == 2).all()  # inner region
    assert (image_segment.mask[12:14, 12:18] == 0).all()  # inner hole
    assert (image_segment.mask[12:18, 12:14] == 0).all()  # inner hole
    assert (image_segment.mask[12:18, 16:18] == 0).all()  # inner hole
    assert (image_segment.mask[16:18, 12:18] == 0).all()  # inner hole
    assert (image_segment.mask[10:12, 10:20] == 1).all()  # outer region
    assert (image_segment.mask[18:20, 10:20] == 1).all()  # outer region
    assert (image_segment.mask[10:20, 10:12] == 1).all()  # outer region
    assert (image_segment.mask[10:20, 18:20] == 1).all()  # outer region




"""A higher level abstraction."""

from typing import Union, Dict
import random
import numpy as np  # type: ignore
import matplotlib.pyplot as plt  # type: ignore

from redbrick.labelset.labelset_base import LabelsetBase
from redbrick.api import RedBrickApi
from redbrick.entity.datapoint import Image, Video
from redbrick.export import ExportImage, ExportVideo
from redbrick.logging import print_info, print_error


class LabelsetLoader(LabelsetBase):
    """Labelset loader class."""

    def __init__(self, org_id: str, label_set_name: str) -> None:
        """Construct Loader."""
        self.org_id = org_id
        self.label_set_name = label_set_name
        self.api_client = RedBrickApi(cache=False)

        print_info("Counting available datapoint...")

        # All datapoints in labelset
        try:
            self.dp_ids, custom_group = self.api_client.get_datapoint_ids(
                self.org_id, self.label_set_name
            )
        except Exception as err:
            print_error(err)
            return

        self.task_type = custom_group.task_type
        self.data_type = custom_group.data_type
        self.taxonomy: Dict[str, int] = custom_group.taxonomy
        print_info("Number of Datapoints = %s" % len(self.dp_ids))

        # Update taxonomy mapper if segmentation
        if self.task_type == "SEGMENTATION":
            self.taxonomy_update_segmentation()

        # Get all users
        try:
            self.users = self.api_client.get_members(self.org_id)
        except Exception as err:
            print_error(err)
            return

    def __getitem__(self, index: int) -> Union[Image, Video]:
        """Get information needed for a single item."""
        dp = self.api_client.get_datapoint(
            self.org_id,
            self.label_set_name,
            self.dp_ids[index],
            self.task_type,
            self.taxonomy,
        )
        return dp

    def export(self, format: str = "redbrick") -> str:
        """Export."""
        if self.data_type == "IMAGE":
            export_img: ExportImage = ExportImage(format=format, labelset=self)
            export_img.export()
            return export_img.cache_dir
        elif self.data_type == "VIDEO":
            export_vid: ExportVideo = ExportVideo(format=format, labelset=self)
            export_vid.export()
            return export_vid.cache_dir
        else:
            err = ValueError(
                "%s data type not supported! Please reach out to \
                    contact@redbrickai.com"
                % self.data_type
            )
            print_error(err)
            return ""

    def number_of_datapoints(self) -> int:
        """Get number of datapoints."""
        return len(self.dp_ids)

    def show_data(self) -> None:
        """Visualize the data."""

        if self.data_type == "VIDEO":
            print_info("Visualizing first 20 frames...")

            num_dps = self.number_of_datapoints()
            if not num_dps:
                return

            idx = random.randint(0, num_dps - 1)

            self[idx].show_data()
            return

        # Image data type
        print_info("Visualizing data and labels...")

        # Prepare figure
        num_dps = self.number_of_datapoints()
        cols = int(np.min([2, num_dps]))
        rows = int(np.min([2, np.ceil(num_dps / cols)]))
        fig = plt.figure()

        # Generate random index list
        list_len = np.min([rows * cols, num_dps])
        indexes = random.sample(range(0, list_len), list_len)

        # Iterate through axes
        for i, idx in enumerate(indexes):

            ax = fig.add_subplot(rows, cols, i + 1)
            self[idx].show_data(ax=ax)  # type: ignore

        plt.tight_layout()
        plt.show()

    def taxonomy_update_segmentation(self) -> None:
        """
        Fix the taxonomy mapper object to be 1-indexed for
        segmentation projects.
        """
        for key in self.taxonomy.keys():
            self.taxonomy[key] += 1
            if self.taxonomy[key] == 0:
                print_error(
                    "Taxonomy class id's must be 0 indexed. \
                        Please contact contact@redbrickai.com for help."
                )
                exit(1)

        # Add a background class for segmentation
        self.taxonomy["background"] = 0



"""
Class for controlling image exports.
"""
from dataclasses import dataclass
from typing import Any, Dict, List
import os
import json
from termcolor import colored
import cv2  # type: ignore
from tqdm import tqdm  # type: ignore
import numpy as np  # type: ignore

from .export_base import ExportBase
from redbrick.logging import print_info


@dataclass
class ExportImage(ExportBase):
    """Handle image exports."""

    def export(self) -> None:
        """Export the images and labels."""
        print_info("Cacheing labels, and data...")
        self.cache()
        print_info("Exported complete in %s" % self.cache_dir)

    def cache(self) -> None:
        """Cache the images and labels."""
        ExportBase.cache(self)

    def cache_bbox(self) -> None:
        """Cache image bounding boxes."""
        # Create YOLO style meta data files
        num_classes = len(list(self.labelset.taxonomy.keys()))
        train_file = "train.txt"
        names_file = "names.txt"
        data_file = "obj.data"
        backup_dir = "backup/"

        # obj.data
        with open(self.cache_dir + "/" + data_file, "w+") as file:
            file.write("classes = %s" % num_classes)
            file.write("\n")
            file.write("train = %s" % ("data" + "/" + train_file))
            file.write("\n")
            file.write("names = %s" % ("data" + "/" + names_file))
            file.write("\n")
            file.write("backup = %s" % backup_dir)

        # names.txt
        with open(self.cache_dir + "/" + names_file, "w+") as file:
            class_names = list(self.labelset.taxonomy.keys())
            for name in class_names:
                file.write(name + "\n")

        # obj_train_data/
        os.mkdir(os.path.join(self.cache_dir, "obj_train_data"))
        taxonomy_mapper = {
            name: idx for idx, name in enumerate(list(self.labelset.taxonomy.keys()))
        }
        image_filepaths = []

        # IMAGE DATA
        for i in tqdm(range(len(self.labelset.dp_ids))):
            dp_ = self.labelset.__getitem__(i)
            dp_entry = {}
            dp_entry["url"] = dp_.image_url_not_signed
            dp_entry["labels"] = []

            # write image data to file
            image_filepath = os.path.join(
                self.cache_dir,
                "obj_train_data",
                str(dp_.image_url_not_signed).replace("/", "_"),
            )
            image_filepaths.append(
                "obj_train_data/" + str(dp_.image_url_not_signed).replace("/", "_")
            )
            cv2.imwrite(  # pylint: disable=no-member
                image_filepath, np.flip(dp_.image_data, axis=2)
            )

            # create the label file name
            filext_idx = image_filepath.rfind(".")
            if not filext_idx:
                filext_idx = -1
            label_filepath = image_filepath[0:filext_idx] + ".txt"

            # write labels to the txt file
            with open(label_filepath, "w+") as file:
                for label in dp_.labels:
                    class_idx = taxonomy_mapper[label.classname[0][-1]]
                    file.write(
                        "%d %.6f %.6f %.6f %.6f \n"
                        % (
                            class_idx,
                            label.xnorm,
                            label.ynorm,
                            label.wnorm,
                            label.hnorm,
                        )
                    )

        # create train.txt file
        with open(os.path.join(self.cache_dir, train_file), "w+") as file:
            for filename in image_filepaths:
                file.write(filename + "\n")

    def cache_segmentation(self) -> None:
        """Cache image segmentation."""
        if self.format == "redbrick":
            # Save masks
            for i in tqdm(range(len(self.labelset.dp_ids))):
                dp_ = self.labelset.__getitem__(i)
                dp_.labels.mask.dump(
                    self.cache_dir
                    + "/"
                    + str(dp_.image_url_not_signed).replace("/", "_")
                    + "__export.dat"
                )

            # Save the class-mapping (taxonomy) in json format in the folder
            with open("%s/class-mapping.json" % self.cache_dir, "w+") as file:
                json.dump(dp_.taxonomy, file, indent=2)

        elif self.format == "redbrick-png":
            # iterate through all labels
            if len(self.labelset.dp_ids) == 0:
                return

            label_info: Dict[Any, Any] = {}
            label_info["labels"] = []
            color_map: Any = None
            for i in tqdm(range(len(self.labelset.dp_ids))):
                dp_ = self.labelset.__getitem__(i)
                colored_mask = dp_.labels.color_mask()
                color_map = dp_.labels.color_map
                # cv2 expected BGR
                # pylint: disable=no-member
                export_url = str(dp_.image_url_not_signed).replace("/", "_")
                cv2.imwrite(
                    os.path.join(self.cache_dir, export_url + "__export.png"),
                    np.flip(colored_mask, axis=2) * 256,
                )
                label_info_entry = {}
                label_info_entry["url"] = dp_.image_url_not_signed
                label_info_entry["createdBy"] = self.labelset.users[dp_.created_by]
                label_info_entry["exportUrl"] = export_url + "__export.png"
                label_info["labels"] += [label_info_entry]

            # Meta-data
            label_info["taxonomy"] = dp_.taxonomy
            label_info["color_map"] = {}
            for key in dp_.taxonomy:
                color_map_ = color_map(dp_.taxonomy[key]).tolist()
                label_info["color_map"][dp_.taxonomy[key]] = color_map_

            # Save the class-mapping (taxonomy) in json format in the folder
            with open("%s/class-mapping.json" % self.cache_dir, "w+") as file:
                json.dump(label_info, file, indent=2)

    def cache_classify(self) -> None:
        """Cache image classification."""
        label_file: List[Any] = []
        for i in tqdm(range(len(self.labelset.dp_ids))):
            dp = self.labelset.__getitem__(i)
            entry = {}
            entry["url"] = dp.image_url_not_signed
            entry["createdBy"] = self.labelset.users[dp.created_by]
            entry["class"] = dp.labels.labels.category
            label_file += [entry]

        with open("%s/export.json" % self.cache_dir, "w+") as file:
            json.dump(label_file, file, indent=2)



"""
Representation for the image datapoint type.
"""
from typing import Union, Any, List
from dataclasses import dataclass, field
import numpy as np  # type: ignore
import matplotlib.pyplot as plt  # type: ignore

from redbrick.entity.label import ImageBoundingBox, ImageClassify, ImageSegmentation
from redbrick.entity.label.segmentation import ImageSegmentationRemoteLabel
from redbrick.entity.label.bbox import ImageBoundingBoxRemoteLabel
from redbrick.logging import print_error
from .base_datapoint import BaseDatapoint


@dataclass
class Image(BaseDatapoint):
    """Image object."""

    dp_id: str
    image_url: str
    image_url_not_signed: str
    image_data: np.ndarray
    labels: Union[ImageBoundingBox, ImageClassify, ImageSegmentation] = field(
        init=False
    )

    def __post_init__(self) -> None:
        """Run after init."""
        if self.task_type == "BBOX":
            # Read in remote labels
            bbox_remote_labels: List[ImageBoundingBoxRemoteLabel] = []
            try:
                bbox_remote_labels = [
                    ImageBoundingBoxRemoteLabel.from_dict(label)
                    for label in self.remote_labels
                ]
            except Exception:
                print_error("Parsing error. Please reach out to contact@redbrickai.com")

            # Create label object
            self.labels = ImageBoundingBox(labels=bbox_remote_labels)

        elif self.task_type == "SEGMENTATION":
            # Read in remote labels
            segment_remote_labels: List[ImageSegmentationRemoteLabel] = []
            try:
                segment_remote_labels = [
                    ImageSegmentationRemoteLabel.from_dict(label)
                    for label in self.remote_labels
                ]
            except Exception as err:
                print_error(err)
                return

            # Create label object
            self.labels = ImageSegmentation(
                remote_labels=segment_remote_labels, classes=self.taxonomy
            )

        elif self.task_type == "CLASSIFY":
            # Read in remote labels
            self.labels = ImageClassify(self.remote_labels)

        else:
            raise ValueError(
                "%s task type is not supported. Please reach out to contact@redbrickai.com."
                % self.task_type
            )

    def show_data(self, ax: Any = None) -> None:
        """Show the data with the ground truth."""
        if ax is None:
            _, ax = plt.subplots()

        # Show the image
        ax.imshow(self.image_data)

        # Render labels
        height, width, _ = self.image_data.shape
        self.labels.show(ax=ax, width=width, height=height)
        title = (
            "..." + self.image_url_not_signed[-10:]
            if len(self.image_url_not_signed) > 10
            else self.image_url_not_signed
        )
        ax.set_title("%s" % title)

        if ax is None:
            plt.show()



"""Testing the image entity."""
import numpy as np  # type: ignore

from redbrick.entity.datapoint import Image
from redbrick.entity.label import (
    ImageBoundingBox,
    ImageBoundingBoxEntry,
    ImageSegmentation,
)
from .bbox_test import LABEL1
from .segmentation_test import remote_label5
from redbrick.entity.label.segmentation import (
    ImageSegmentationRemoteLabel,
    ImageSegmentation,
)
import matplotlib.pyplot as plt

IMAGE = Image(
    org_id="123",
    label_set_name="abc",
    taxonomy={"person": 2, "car": 1},
    task_type="BBOX",
    remote_labels=LABEL1,
    dp_id="123",
    image_url="abc.png",
    image_url_not_signed="abc234.png",
    image_data=np.zeros((10, 10, 3)),
    created_by="123",
)

IMAGE_SEGMENT = Image(
    org_id="123",
    label_set_name="abc",
    taxonomy={"person": 2, "car": 1},
    task_type="SEGMENTATION",
    remote_labels=remote_label5,
    dp_id="123",
    image_url="abc.png",
    image_url_not_signed="abc234.png",
    image_data=np.zeros((100, 100, 3)),
    created_by="123",
)


def test_init_bbox() -> None:
    """Test basic init."""
    # arrange
    image = IMAGE

    # assert
    assert isinstance(image.labels, ImageBoundingBox)
    assert len(image.labels.labels) == 1
    assert isinstance(image.labels.labels[0], ImageBoundingBoxEntry)


def test_init_segmentation() -> None:
    """Test basic init of segmentation."""
    # arrange
    image = IMAGE_SEGMENT

    # assert
    assert isinstance(image.labels, ImageSegmentation)
    assert len(image.remote_labels) == 2




"""
Representation of a segmentation label.
"""

from dataclasses import dataclass, field
from typing import Dict, List, Any, Optional
import copy
import numpy as np  # type: ignore
from skimage import draw  # type: ignore
import matplotlib.cm  # type: ignore
import matplotlib.pyplot as plt


@dataclass
class Pixel:
    """Representation of pixel field in RemoteLabel."""

    imagesize: List[int]
    regions: List[List[List[int]]]
    holes: List[List[List[int]]]


@dataclass
class ImageSegmentationRemoteLabel:
    """Single Remote label object for image segmentation."""

    category: List[List[str]]
    pixel: Pixel
    labelid: str

    @classmethod
    def from_dict(cls, obj: Dict[Any, Any]) -> "ImageSegmentationRemoteLabel":
        """Create a remote label object from dict."""
        pixel = Pixel(
            imagesize=obj["pixel"]["imagesize"],
            regions=obj["pixel"]["regions"],
            holes=obj["pixel"]["holes"],
        )
        return cls(category=obj["category"], pixel=pixel, labelid=obj["labelid"],)


@dataclass
class ImageSegmentation:
    """Image segmentation object."""

    classes: Dict[str, int]
    remote_labels: List[ImageSegmentationRemoteLabel]

    url: str = field(init=False)
    mask: np.ndarray = field(init=False)

    # Matplotlib CMAP returns color code
    # RGB 0-256 numpy array of int's
    color_map: Any = lambda idx: (
        np.array(matplotlib.cm.get_cmap("tab10")(idx))[0:3] * 256
    ).astype(int)

    def __post_init__(self) -> None:
        """
        Convert remote label format to mask format to be stored.
        """
        # TODO: Check this code for scenario with no label
        imagesize = self.remote_labels[0].pixel.imagesize
        mask = np.zeros([imagesize[1], imagesize[0]])

        for label_obj in self.remote_labels:
            label = label_obj.pixel
            category_name = label_obj.category[0][-1]
            class_id = self.classes[category_name]
            regions = copy.deepcopy(label.regions)
            holes = copy.deepcopy(label.holes)
            imagesize = label.imagesize

            # Iterate through the regions and create mask
            mask_ = np.zeros([imagesize[1], imagesize[0]])
            if regions and len(regions) > 0:
                for region in regions:
                    if len(np.array(region).shape) == 1:
                        # Don't add empty regions to the mask
                        break

                    # [x,y] needs to correspond to [r,c]
                    region = np.flip(region, axis=1)

                    mask__ = (
                        draw.polygon2mask([imagesize[1], imagesize[0]], region).astype(
                            float
                        )
                        * class_id
                    )

                    # Add on new regions to root mask object
                    # class_id_indexes = np.where(mask__ == class_id)
                    # mask_[class_id_indexes] = class_id
                    mask_ += mask__

            # Iterate through the holes and create the negative mask
            neg_mask = np.zeros([imagesize[1], imagesize[0]])
            if holes and len(holes) > 0:
                for hole in holes:
                    if len(np.array(hole).shape) == 1:
                        # Don't add empty hole to the negative mask
                        break

                    # [x,y] needs to correspond to [r,c]
                    hole = np.flip(hole, axis=1)

                    neg_mask_ = (
                        draw.polygon2mask([imagesize[1], imagesize[0]], hole).astype(
                            float
                        )
                        * class_id
                    )

                    # Add on new holes to negative mask
                    neg_mask += neg_mask_

            # create the holes for this label object,
            # and remove any negative values that arise from holes
            # "spilling outside" the region
            mask_ -= neg_mask
            neg_idxs = np.where(mask_ < 0)
            mask_[neg_idxs] = 0

            # overlapping regions would have gotten added together
            # set the overlapped regions back to the classid.
            overlap_indexes = np.where(mask_ > class_id)
            mask_[overlap_indexes] = class_id

            # Merge
            class_idx_not_zero = np.where(mask_ != 0)
            mask[class_idx_not_zero] = mask_[class_idx_not_zero]

        # Subtrack out the holes from the region mask
        self.mask = mask

    def color_mask(self) -> np.ndarray:
        """Return a RGB colored mask."""
        mask_class_ids = np.unique(self.mask)
        color_mask = np.zeros([self.mask.shape[0], self.mask.shape[1], 3])

        # Loop through class range
        for id_ in mask_class_ids:
            if id_ == 0:
                # Do not color class id 0 i.e. the background
                continue

            pixel_class_index = np.where(self.mask == id_)

            # generate colors
            color_mask[pixel_class_index] = self.color_map(int(id_))

        color_mask /= 256
        return color_mask

    def show(
        self, ax: Any = None, width: Optional[int] = None, height: Optional[int] = None
    ) -> None:
        """Show the segment labels."""
        mask = self.color_mask()
        ax.imshow(mask, alpha=0.3)
